/*
    Copyright (C) 2007 by Michael Dale Long
  
    This file is part of DuSTED.
  
    DuSTED is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
  
    DuSTED is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with DuSTED.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <nds.h>
#include <string.h>
#include "dusted.h"
#include "canvas.h"
#include "palette.h"
#include "undo.h"
#include "engine.h"
#include "tilewin.h"

WIDGET_IMPL(Canvas, CID_CANVAS);

#define PIXEL_INDEX(x, y) ((y) * CANVAS_WIDTH + (x))

void Canvas::Init(void) {
	// Start with the first tile
	mCurTile = 0;
	// Let the canvas render
	mVisible = true;
	// Read in the pixels from the PBuffer
	UpdatePixels();
}


void Canvas::SetArea(int x, int y, int w, int h) {
	// Set the area size
	mArea.Set(x, y, w, h);
	// Set the number of real pixels per canvas pixel
	mTilePixelSize.Set(0, 0, mArea.w / 8, mArea.h / 8);
	// Setup tile used for rendering
	mRenderTile = mTilePixelSize;
	// Render all pixels on the first frame
	Refresh();
}


void Canvas::HandleTouch(uint32 state, int16 x, int16 y) {
	// Check if the touch pen is down and inside our area
	if (state & PS_DOWN && mArea.ContainsPoint(x, y)) {
		// Set offset to 0, 0 relative to the canvas origin
		x -= mArea.x;
		y -= mArea.y;

		// Find the x, y pixel position
		x /= mTilePixelSize.w;
		y /= mTilePixelSize.h;

		// See if we haven't already been set to this color
		int pixIndex = PIXEL_INDEX(x, y);
		
		TileEditorWindow *win = TileEditorWindow::FromWidget(mParent);
		LogAssert(win == NULL, "Canvas::mParent invalid!\n");
		if (mPixels[pixIndex] != win->mPaletteWidget->mPalIndex) {
			// Setup the action 
			Action action;
			action.mCode = AC_COLOR;
			action.mColor.mCanvas = this;
			action.mColor.mOldIndex = mPixels[pixIndex];
			action.mColor.mX = x;
			action.mColor.mY = y;
			action.mColor.mNewIndex = win->mPaletteWidget->mPalIndex;

			// Set the pixel color to the current palette color
			win->mActions.Perform(action);
			win->UpdateSubView();
		}
	}
}


void Canvas::Render(void) {
	// See if a tile has changed
	if (mPoints.Size() > 0) {
		register uint32 size = mPoints.Size();
		for (register uint32 i = 0; i < size; i++) {
			// Setup the render tile origin
			mRenderTile.x = mArea.x + mPoints[i].x * mTilePixelSize.w;
			mRenderTile.y = mArea.y + mPoints[i].y * mTilePixelSize.h;
			// Render the tile
			mRenderTile.DrawFilled(reinterpret_cast<TileEditorWindow *> (mParent)->mPaletteWidget->mPalette[mPixels[PIXEL_INDEX(mPoints[i].x, mPoints[i].y)]], VRAM_A);
		}
		// Clear the list, don't free memory associated with the list
		mPoints.RemoveAll();
	}
}


void Canvas::Refresh(void) {
	// Clear the mPoints list and add all points to it so they'll all be rendered next vblank
	mPoints.RemoveAll();
	for (int x = 0; x < 8; x++) {
		for (int y = 0; y < 8; y++) {
			mPoints.Append(Point(x, y));
		}
	}
}


void Canvas::SetColor(uint16 x, uint16 y, uint16 color) {
	if (x >= CANVAS_WIDTH || y >= CANVAS_HEIGHT) {
		LogFatal("Canvas given bad coords (%u, %u)\n", x, y);
	}

	mPixels[PIXEL_INDEX(x, y)] = color;

	// Note that the tile changed so it will be rendered
	mPoints.AppendOnce(Point(x, y));
}


void Canvas::UpdatePixels(void) {
	// Pixels reside in the PBuffer, after the header (64) and the palette (512).
	mPixels = (uint8 *) (gEngine->mPBuffer + (576 + mCurTile * 64));
}


void Canvas::PrevTile(void) {
	mCurTile--;
	if (mCurTile < 0) {
		mCurTile = MAX_TILES - 1;
	}
	UpdatePixels();
	Refresh();
}


void Canvas::NextTile(void) {
	mCurTile++;
	if (mCurTile >= MAX_TILES) {
		mCurTile = 0;
	}
	UpdatePixels();
	Refresh();
}


void Canvas::SetTile(int index) {
	mCurTile = index;
	// Wrap out of bound indices
	if (mCurTile < 0) {
		mCurTile += MAX_TILES;
	}
	if (mCurTile >= MAX_TILES) {
		mCurTile -= MAX_TILES;
	}
	UpdatePixels();
	Refresh();
}

